import numpy as np
import re

second_conversions = {"s": 1, "m": 60, "h": 3600, "d": 86400, "w": 604800, "y": 31536000}


def convert_string_to_seconds(s):
    try:
        _, n, uom = re.split("(\d+)", s)
        n = int(n)
        if uom not in second_conversions.keys():
            raise ValueError(
                f"{uom} is not valid in the list of valid values: {second_conversions.keys()}"
            )
        return n * second_conversions[uom]
    except:
        raise ValueError(f"Invalid format.  Expecting NumberUOM (15h) but instead recieved {s}")


class TimeIndex(np.ndarray):
    """ 
    TimeIndex
    ----------

    This class converts a a float array to a np.datetime64 array
    """

    _dt = None

    def __new__(self, data):
        """Creates a new array based on the based on a float array"""
        obj = np.array(data, dtype=np.float).view(self)
        obj._dt = None
        return obj

    def __array_finalize__(self, obj):
        """ Finalize the array """
        if obj is None:
            return

    @property
    def dt(self):
        """ creates a datetime object for the time index """
        if not self._dt:
            vfunc = np.vectorize(lambda x: np.datetime64(int(x), "s"))
            self._dt = vfunc(self)
        return self._dt



import numpy as np


class TimeSeriesCollection:
    """ A grouping of rows and columns for a TimeSeries 
    
    Attributes
    ----------

    data : np.array
        The place holder for array to be returned
    axis : int
        The axis for the function to be applied on
    collapse_index : bool
        Should the rows be squashed into 1 row
    collapse_time_index : int
        Which row index should be used.  0 for first, -1 for last
    chunks : [TimeChunk]
        A list of chunks to iterate over
    parent : TimeSeries
        The parent of the collection
    init_data : np.array
        If there is any initial data to start with
    """

    data = None
    axis = 0
    collapse_index = False
    collapse_time_index = 0
    chunks = []
    parent = None
    init_data = None

    def __getattr__(self, attr_name):
        """ Return a method for process each chunk """
        from .timeseries import TimeSeries

        def method(*args, **kwargs):
            if attr_name in dir(np) and callable(getattr(np, attr_name)):
                rtn_time_series = self.init_data
                for chunk in self.chunks:
                    if "axis" not in kwargs:
                        kwargs["axis"] = self.axis
                    if isinstance(getattr(np, attr_name), np.ufunc):
                        data = getattr(np, attr_name).reduce(
                            self.parent[chunk.row_mask][:, chunk.col_mask], *args, **kwargs
                        )
                    else:
                        data = getattr(np, attr_name)(
                            self.parent[chunk.row_mask][:, chunk.col_mask], *args, **kwargs
                        )
                    if not data.shape:  # not an array
                        return data
                    if len(data.shape) == 1:
                        if kwargs["axis"] == 0:
                            data = data.reshape((1, data.shape[0]))
                    elif data.shape[1] == 1 and data.shape[0] != 1:
                        if kwargs["axis"] == 0:
                            data = data.reshape((data.shape[1], data.shape[0]))

                    row_count = data.shape[0]
                    if len(data.shape) == 1:
                        col_count = 1
                    else:
                        col_count = data.shape[1]

                    time_index = self.parent.time[chunk.row_mask]
                    if self.collapse_index and row_count > 0:
                        time_index = [time_index[self.collapse_time_index]]
                        row_count = 1
                    new_ts = TimeSeries(
                        shape=(row_count, col_count),
                        time=time_index,
                        labels=self.parent._least_common_labels(chunk.col_mask),
                    )
                    new_ts[:] = data
                    if rtn_time_series is None:
                        rtn_time_series = new_ts
                    else:
                        if self.collapse_index:  # need to vstack this
                            rtn_time_series.vstack(new_ts)
                        else:
                            rtn_time_series = rtn_time_series.merge([new_ts])
                return rtn_time_series
            else:
                raise AttributeError

        return method



import numpy as np


class NumpyArray(np.ndarray):
    """
    This class extends numpy.array to be used with the TimeSeries class

    """

    def __new__(self, shape, parent):
        """
        Create a new array

        Parameters
        ----------
        shape : tuple
            the size to allocate for the array
        parent : TimeSeries
            the TimeSeries object that owns this array
        """
        obj = np.empty(shape, dtype=np.float64).view(self)
        obj[:] = np.nan
        obj.parent = parent
        return obj

    def __array_finalize__(self, obj):
        """Used to add labels to the object"""
        if obj is None:
            return
        self.parent = getattr(obj, "parent", None)

    def __array_wrap__(self, out_arr, context=None):
        """ Wrap the array as a TimeSeries object after processing """
        return self.parent.wrap_new_data(out_arr)




import numpy as np
import csv
from itertools import compress
from timeseriesql.query import Query
from timeseriesql.timeseries import TimeSeries
from timeseriesql.ast import Metric, Value


def convert_to_float(s):
    try:
        return float(s)
    except:
        return np.nan


def convert_date_to_float(s):
    try:
        return float(s)
    except:
        return np.datetime64(s).astype(float)


class CSVFile:
    def __init__(self, name):
        self.name = name
        self.filters = []


class CSVBackend(Query):

    _labels = None
    _noheader = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args)

    def labels(self, labels, noheader=False):
        self._labels = labels
        self._noheader = noheader
        return self

    def _header_mask(self, filters, row):
        mask = [0] + [1 for x in row[1:]]
        if len(filters) > 0:
            for f in filters:
                if not any(mask):
                    break
                for i, x in enumerate(row[1:], 1):
                    if mask[i] == 0:
                        continue
                    if f["op"] == "==":
                        if x != f["right"]:
                            mask[i] = 0
                    elif f["op"] == "!=":
                        if x == f["right"]:
                            mask[i] = 0
                    elif f["op"] == "in":
                        if x not in f["right"]:
                            mask[i] = 0
                    elif f["op"] == "not in":
                        if x in f["right"]:
                            mask[i] = 0
                    else:
                        raise NotImplementedError(
                            f"unsupported op type of '{f['op']}' for the filter"
                        )
        return mask

    def compareequal(self, left, right):
        return {"op": "==", "right": right}

    def comparenotequal(self, left, right):
        return {"op": "!=", "right": right}

    def comparein(self, left, right):
        return {"op": "in", "right": right}

    def comparenotin(self, left, right):
        return {"op": "not in", "right": right}

    def filter(self, left, right):
        left.filters.append(right)
        return left

    def traverse_tree(self, root):
        if root:
            if isinstance(root, Metric):
                return CSVFile(root.value)
            if isinstance(root, Value):
                return root.value
            else:
                left = self.traverse_tree(root.left)
                right = self.traverse_tree(root.right)
                try:
                    op = self.__getattribute__(root.__class__.__name__.lower())
                    return op(left, right)
                except Exception as e:
                    raise NotImplementedError(
                        f"AST class of {root.__class__.__name__} is not supported"
                    )

    def execute_plan(self):
        ast = self._generate_plan()
        csvobject = self.traverse_tree(ast)
        labels = []
        data = []
        time = []
        with open(csvobject.name) as csvfile:
            filereader = csv.reader(csvfile)
            read_header = False
            mask = None
            for row in filereader:
                if not read_header:
                    if not self._noheader:
                        read_header = True
                        if not self._labels:
                            mask = self._header_mask(csvobject.filters, row)
                            labels = [{"label": x} for x in compress(row, mask)]
                        else:
                            mask = [0] + [1 for x in row[1:]]
                            labels = self._labels
                        continue
                    else:
                        mask = [0] + [1 for x in row[1:]]
                        labels = self._labels
                        read_header = True
                data.append(list(map(convert_to_float, compress(row, mask))))  # row zero is time
                time.append(row[0])
        t = TimeSeries(
            shape=(len(data), len(labels)),
            time=list(map(convert_date_to_float, time)),
            labels=labels,
        )
        t[:] = data
        return t




import numpy as np
import numbers
from collections.abc import Sequence
from .np_array import NumpyArray
from .time import TimeIndex, convert_string_to_seconds
from itertools import compress
from .timeseries_collection import TimeSeriesCollection
from .time_chunk import TimeChunk
from .utils import dispatchmethod


class TimeSeries:
    """ The base object to work with time series 
    
    Attributes
    ----------
    data : NumpyArray
        The data associated with the time series object
    _time : np.array
        Then shared time index for the time series
    labels : [dict]
        A list of key value pairs for the streams
    """
    def __init__(self, *args, shape=(4, 3), labels=None, time=[], **kwargs):
        """
        Initialize the object

        Parameters
        ----------
        shape : tuple
            the shape to use to make the data array
        labels : [dict]
            a list of key value pairs for the streams
        time : array like
            a list of time indicies
        """
        self.data = NumpyArray(shape, self)
        if isinstance(labels, list):
            self.labels = [l for l in labels]
        elif labels:
            self.labels = [labels]
        else:
            self.labels = []
        if time != []:
            time = time[-self.data.shape[0]:]
        self._time = np.array(time, dtype=np.float64)

    def __array__(self):# pragma: no cover
        return self.data

    def __array_wrap__(self, out_arr, context=None):
        return self.wrap_new_data(out_arr)

    def __str__(self):# pragma: no cover
        """ String override which gives a larger amount of details """
        labels = ""
        for i, l in enumerate(self.labels):
            if i > 4:
                labels += f"{len(self.labels) - 5} more....\n"
                break
            for k, v in l.items():
                labels += f"{k}:{v} "
            labels += "\n"
        if labels == "":
            labels = None
        time_range = ""
        if self.time[0] > 0:
            time_range = f"Time Range\n===========\n{self.time.dt[0]} ... {self.time.dt[-1]}\n\n"
        return f"\nLabels:\n===========\n{labels}\n{time_range}Data:\n============\n{self.data.__str__()}"

    def __getattr__(self, attr_name):
        """ Attach the data attributes to the top level """
        if attr_name in dir(self.data) and not callable(getattr(self.data, attr_name)):
            return getattr(self.data, attr_name)

        def method(*args, **kwargs):
            if attr_name in dir(self.data):
                return getattr(self.data, attr_name)(*args, **kwargs)
            if attr_name in dir(np) and callable(getattr(np, attr_name)):
                return getattr(np, attr_name)(self, *args, **kwargs)
            else:
                raise AttributeError

        return method

    def __getitem__(self, items):
        """Add functionality for filtering on labels and datetime objects"""
        # if a label filter, return view for the correct streams
        try:
            time_index = self._convert_time_index(items)
            return self.wrap_new_data(self.data.__getitem__(items), time=self.time[time_index], ndx=items)
        except IndexError as e:
            index = self._handle_index_error(items, e)
            time_index = self._convert_time_index(index)
            return self.wrap_new_data(self.data.__getitem__(index), time=self.time[time_index], ndx=index)
        except TypeError as e:
            index = self._handle_type_error(items, e)
            time_index = self._convert_time_index(index)
            return self.wrap_new_data(self.data.__getitem__(index), time=self.time[time_index], ndx=index)

    def __setitem__(self, key, item):
        """ Set the item based off of a key in data """
        try:
            self.data.__setitem__(key, item)
        except IndexError as e:
            return self.data.__setitem__(self._handle_index_error(key, e), item)
        except TypeError as e:
            return self.data.__setitem__(self._handle_type_error(key, e), item)

    @dispatchmethod
    def _handle_index_error(self, key, error):
        l = len(key)
        if l != 1:
            raise error
        if l == 1:
            filters = self.argfilter(key)
            if len(filters) < 1:
                return slice(0,0)  # maybe should return something else?
            return (slice(None), filters)

    @_handle_index_error.register(int)
    def _(self, key, error):
        raise error

    @_handle_index_error.register(tuple)
    @_handle_index_error.register(slice)
    @_handle_index_error.register(np.datetime64)
    def _(self, key, error):
        time_slice = key
        extra_slices = None
        if isinstance(key, tuple):
            time_slice, extra_slices = key
        start, stop = self._get_slice_by_datetime(time_slice)
        if (start, stop) == (None, None):
            raise error
        if extra_slices:
            return (start, extra_slices)
        return start

    def _handle_type_error(self, key, error, item = None):
        # build new slice
        time_slice = key
        extra_slices = None
        if isinstance(key, tuple):
            time_slice, extra_slices = key
        if isinstance(time_slice, slice):
            start, stop = self._get_slice_by_datetime(time_slice.start, time_slice.stop)
            if (start, stop) == (None, None):
                raise error
            if extra_slices:
                return (slice(start, stop), extra_slices)
            else:
                return slice(start, stop)
        raise error

    def _convert_time_index(self, index):
        if not isinstance(index, Sequence):
            return index
        else:
            return index[0]

    def _merge_time_indicies(self, t1, t2):
        """Merge time indicies to crate an a new index of the union of the two indicies"""
        res = sorted(set(t1).union(set(t2)))
        return np.array(res)

    def _get_group_masks(self, group):
        """Return a mask for each label"""
        if not isinstance(group, list):
            group = [group]
        groupings = {}
        for i, stream in enumerate(self.labels.copy()):
            key = ""
            for g in group:
                if g in stream:
                    key += stream[g]
            if key in groupings:
                groupings[key].append(i)
            else:
                groupings[key] = [i]
        return [x for x in groupings.values()]

    def _get_slice_by_datetime(self, start_date=None, stop_date=None):
        """Return the indicies for a TimeSeries object searching by a datetime object"""
        start, stop = (start_date, stop_date)
        if isinstance(start_date, np.timedelta64):
            start_date = self.time.dt[0] + start_date
        if isinstance(start_date, np.datetime64):
            start = np.argwhere(self.time >= start_date.astype(float))
            if len(start) > 0:
                start = start[0][0]
            else:
                raise IndexError("No timestamps exist at or after %s" % start_date)
        if isinstance(stop_date, np.timedelta64):
            if start_date:
                stop_date = start_date + stop_date
            else:
                stop_date = self.time.dt[-1] - stop_date
        if isinstance(stop_date, np.datetime64):
            stop = np.argwhere(self.time >= stop_date.astype(float))
            if len(stop) > 0:
                stop = stop[0][0]
            else:
                stop = np.argwhere(self.time <= stop_date.astype(float))
                if len(stop) > 0:
                    stop = None
                else:
                    raise IndexError("No timestamps exist before or at %s" % stop_date)
        return (start, stop)

    def _least_common_labels(self, mask):
        """ create true false mask - probably a better way to do this """
        label_mask = []
        next_i = 0
        if isinstance(mask, slice):
            mask = [x for x in range(0,self.data.shape[1])]
        for m in mask:
            if next_i == m:
                label_mask.append(1)
                next_i += 1
            else:
                for _ in range(next_i,m):
                    label_mask.append(0)
                label_mask.append(1)
                next_i = m + 1
        labels = list(compress(self.labels.copy(), label_mask))
        common_labels = []
        if labels:
            common_labels = labels[0].copy()
            if len(labels) > 1:
                for l in labels[1:].copy():
                    for k in common_labels.copy().keys():
                        if k not in l or l[k] != common_labels[k]:
                            del common_labels[k]
        return common_labels

    def _get_unique_keys(self):
        """Get list of unique keys"""
        labels = self.labels[0].copy()
        for l in self.labels[1:]:
            for k in labels.copy().keys():
                if k not in l or l[k] != labels[k]:
                    del labels[k]
        return labels
    
    @property
    def time(self):
        """Return only the TimeSeries index"""
        return TimeIndex(self._time.view(np.ndarray))

    def wrap_new_data(self, out_arr, time=None, ndx=None):
        """ Wrap a value returned from a ufunc in a TimeSeries object """
        if not isinstance(out_arr, np.ndarray) or not out_arr.shape:
            return out_arr
        if time is None:
            time = self.time
        if not isinstance(time, np.ndarray):
            time = [time]
        if self.data.shape == out_arr.shape:
            new_to = TimeSeries(shape=out_arr.shape, time=time, labels=self.labels)
            new_to.data[:] = out_arr
        elif self.data.shape[0] == out_arr.shape[0]:
            rows = self.data.shape[0]
            columns = out_arr.shape[1] if len(out_arr.shape) > 1 else 1
            out_arr = out_arr.reshape((rows, columns))
            new_to = TimeSeries(shape=(rows, columns), time=time, labels=self.labels)
            new_to.data[:] = out_arr
            if ndx and len(ndx) > 1 and isinstance(ndx[1], (slice, int, list)):
                if isinstance(ndx[1], list):
                    labels = [self.labels[x] for x in ndx[1]]
                else:
                    labels = self.labels[ndx[1]]
                if not isinstance(labels, list):
                    labels = [labels]
                new_to.labels = labels
            else:
                label = self.parent._least_common_labels([x for x in range(0,self.parent.data.shape[1])])
                new_to.labels = [label for _ in range(columns)]
        elif self.data.shape[1] == out_arr.shape[0]:
            columns = self.data.shape[1]
            rows = out_arr.shape[0] if len(out_arr.shape) > 1 else 1
            if out_arr.size % (rows * columns) != 0: #not broadcastable
                new_to = TimeSeries(shape=out_arr.shape, time=time, labels=self.labels)
            else:
                out_arr = out_arr.reshape((rows, columns))
                new_to = TimeSeries(shape=(rows, columns), time=time, labels=self.labels)
            new_to.data[:] = out_arr
        else:
            new_to = TimeSeries(shape=out_arr.shape, time=time, labels=self.labels)
            new_to.data[:] = out_arr
            #return out_arr
        return new_to

    def vstack(self, new_data):
        """ Vertically stack data to a TimeSeries """
        data = NumpyArray(shape=(self.shape[0] + new_data.shape[0], new_data.shape[1]), parent=self)
        data[0:self.shape[0],:] = self.data
        data[self.shape[0]:,:] = new_data.data
        self.data = data
        #concat time index
        self._time = np.array(np.concatenate([self._time, new_data._time]), dtype=np.float64)


    def merge(self, tseries, fill=np.nan):
        """Merge multiple timeseries into one TimeSeries object

        Parameters
        ----------
        tseries: list of TimeSeries
            a list with time series that need to be merged with self
        fill:
            can be substituted with a float or np.float if nan is not desired for empty columns

        Returns
        -------
        A new TimeSeries object

        Example
        -------
        t1.merge([t2, t3])
        >>> TimeSeries([[1.56563280e+09, 5.22240746e-01, 1.93171146e+01, ...,
             5.38888889e-01, 9.68305908e+00, 2.98059849e+00],
            [1.56563370e+09, 5.33333333e-01, 1.95934302e+01, ...,
             4.55587586e-01, 9.99463997e+00, 3.07007493e+00],
            [1.56563460e+09, 5.22222222e-01, 1.94288035e+01, ...,
             4.55471427e-01, 7.87254330e+00, 2.96265267e+00],
            ...,
            [1.56571650e+09, 5.27796301e-01, 1.96337589e+01, ...,
             1.35845181e+00, 9.80688437e+00, 3.09112516e+00],
            [1.56571740e+09, 5.27705227e-01, 1.94194939e+01, ...,
             1.28194444e+00, 8.31783752e+00, 3.09129471e+00],
            [1.56571830e+09, 5.83333333e-01, 2.04479746e+01, ...,
             8.54166667e-01, 7.78839232e+00, 3.19808880e+00]])
        """
        # find out if timeindex are the same
        timeindex = self.time
        columns = self.shape[1]
        if not isinstance(tseries, list):
            tseries = [tseries]
        for t in tseries:
            if not isinstance(t, TimeSeries):
                raise (
                    NotImplementedError(
                        "A object of type {} cannot be combined with a TimeSeries".format(type(t))
                    )
                )
            if not np.array_equal(timeindex, t.time):
                # merge time index
                timeindex = self._merge_time_indicies(timeindex, t.time)
            columns += t.shape[1]

        # create new series
        new_t = TimeSeries(
            shape=(len(timeindex), columns), labels=self.labels.copy(), time=timeindex
        )
        if  fill != np.nan and isinstance(fill, (float, np.float64)):
            new_t[:] = fill

        row_mask = np.isin(timeindex, self.time)
        new_t.data[row_mask, 0 : self.data.shape[1]] = self.data

        # set the counter
        column_counter = self.shape[1]
        for t in tseries:
            t_columns = t.shape[1] + column_counter
            row_mask = np.isin(timeindex, t.time)
            new_t[row_mask, column_counter:t_columns] = t.data
            column_counter = t_columns
            for l in t.labels:
                new_t.labels.append(l.copy())
        return new_t

    def argfilter(self, clauses):
        """Filter the columns of data based on labels.

        Parameters
        ----------
        clauses : list
            clauses to filter on

        Returns
        -------
        a list of indexes for matching columns

        Example
        -------
        t.argfilter([{'hostname': 'host2'}])
        >>> [1,4,6]
        """
        stream_indexes = []
        for i, l in enumerate(self.labels.copy()):
            for k, v in clauses.items():
                if not isinstance(v, list):
                    v = [v]
                if k in l.keys() and l[k] in v:
                    stream_indexes.append(i)  # because first column is the time index
                    continue
        return stream_indexes

    def filter(self, clauses):
        """Filter the columns of data based on labels.

        Parameters
        ----------
        clauses : list
            clauses to filter on

        Returns
        -------
        a np.array view for matching columns

        Example
        -------
        t[{'hostname': 'host2'}]
        >>> TimeSeries([[1.56563280e+09, 5.22240746e-01, 1.93171146e+01, ...,
             5.38888889e-01, 9.68305908e+00, 2.98059849e+00],
            [1.56563370e+09, 5.33333333e-01, 1.95934302e+01, ...,
             4.55587586e-01, 9.99463997e+00, 3.07007493e+00],
            [1.56563460e+09, 5.22222222e-01, 1.94288035e+01, ...,
             4.55471427e-01, 7.87254330e+00, 2.96265267e+00],
            ...,
            [1.56571650e+09, 5.27796301e-01, 1.96337589e+01, ...,
             1.35845181e+00, 9.80688437e+00, 3.09112516e+00],
            [1.56571740e+09, 5.27705227e-01, 1.94194939e+01, ...,
             1.28194444e+00, 8.31783752e+00, 3.09129471e+00],
            [1.56571830e+09, 5.83333333e-01, 2.04479746e+01, ...,
             8.54166667e-01, 7.78839232e+00, 3.19808880e+00]])
        """
        return self[[self.argfilter(clauses)]]

    def group(self, group):
        """ Create a collection based on a grouping """
        masks = self._get_group_masks(group)
        collection = TimeSeriesCollection()
        collection.chunks = (TimeChunk(slice(None, None, None), mask) for mask in masks)
        collection.axis = 1
        collection.parent = self
        return collection

    def resample(self, period):
        """ Create a collection based on resampling by time """
        if isinstance(period, str):
            period = convert_string_to_seconds(period)
        collection = TimeSeriesCollection()
        collection.axis = 0
        collection.parent = self
        collection.collapse_index = True
        columns = [x for x in range(0,self.data.shape[1])]
        collection.chunks = (TimeChunk(slice(*self._get_slice_by_datetime(np.datetime64(int(t), "s"), np.timedelta64(period, "s"))), columns) for t in range(int(self.time[0]), int(self.time[-1]), period)) # pylint: disable-msg=too-many-function-args
        return collection

    def to_pandas(self):
        """ Return two pandas objects.  One for the label data and one for the measurements"""

        import pandas as pd

        columns = ["stream_" + str(x) for x in range(len(self.labels))]
        data = pd.DataFrame(data=self.data, columns=columns)
        data["time_index"] = pd.to_datetime(self.time, unit="s")
        data.set_index("time_index")
        label_columns = list(set(key for dic in self.labels for key in dic.keys()))
        label_data = [[row.get(key, None) for row in self.labels] for key in label_columns]
        labels = pd.DataFrame(columns=label_columns)
        for i, k in enumerate(label_columns):
            labels[k] = label_data[i]
        return (data, labels)

    def rolling_window(self, size, stepsize=1):
        """ Create a collection based on a rolling window """
        collection = TimeSeriesCollection()
        collection.axis = 0
        collection.parent = self
        collection.collapse_index = True
        init_data = TimeSeries(shape=(size-1, self.shape[1]), labels=self.labels, time=self._time[:size-1])
        init_data[:] = np.nan
        collection.init_data = init_data
        collection.collapse_time_index = -1
        collection.chunks = (TimeChunk(slice(i, i+size), slice(None, None, None)) for i in range(0, len(self)-size+1, stepsize))
        return collection

    def fillnan(self, method):
        """ Replace np.nan with a number, or foward/back fill """
        if isinstance(method, str) and method in ['bfill', 'ffill']:
            modifier = -1
            if method == 'bfill':
                modifier = 1
            for i in range(self.data.shape[1]):
                while True:
                    idx = np.argwhere(np.isnan(self.data[:,i]))
                    if len(idx) == 0:
                        break
                    row, col = idx[-1]
                    col = i
                    if 0 < (row + modifier) < len(self.data):
                        self.data[row,col] = self.data[row+modifier,col]
                    else:
                        self.data[row,col] = 0
        elif isinstance(method, numbers.Number):
            self.data[np.isnan(self.data).astype(bool)] = method #need as bool because all values are floats
        else:
            raise ValueError("Method should be 'bfill','ffill', or a number")


    def bfill(self):
        """ Back fill np.nan """
        return self.fillnan('bfill')

    def ffill(self):
        """ Forward fill np.nan """
        return self.fillnan('ffill')

    def fill(self, value):
        """ fill with a number np.nan """
        return self.fillnan(value)

    def diff(self, *args, **kwargs):
        """ Find the n difference for rows """
        if 'axis' not in kwargs:
            kwargs['axis'] = 0
        data = np.diff(self, *args, **kwargs)
        new_t = TimeSeries(shape=data.shape, time=self.time[-data.shape[0]:])
        new_t.labels = self.labels
        new_t[:] = data
        return new_t

    def copy(self):
        """Override the copy method to include the labels

        Parameters
        ----------
        None

        Returns
        -------
        A new TimeSeries object

        Example
        -------
        new_t.copy()
        >>> TimeSeries([[1.56563280e+09, 5.22240746e-01, 1.93171146e+01, ...,
             5.38888889e-01, 9.68305908e+00, 2.98059849e+00],
            [1.56563370e+09, 5.33333333e-01, 1.95934302e+01, ...,
             4.55587586e-01, 9.99463997e+00, 3.07007493e+00],
            [1.56563460e+09, 5.22222222e-01, 1.94288035e+01, ...,
             4.55471427e-01, 7.87254330e+00, 2.96265267e+00],
            ...,
            [1.56571650e+09, 5.27796301e-01, 1.96337589e+01, ...,
             1.35845181e+00, 9.80688437e+00, 3.09112516e+00],
            [1.56571740e+09, 5.27705227e-01, 1.94194939e+01, ...,
             1.28194444e+00, 8.31783752e+00, 3.09129471e+00],
            [1.56571830e+09, 5.83333333e-01, 2.04479746e+01, ...,
             8.54166667e-01, 7.78839232e+00, 3.19808880e+00]])
        """
        ts = TimeSeries(shape=self.shape, labels=self.labels.copy(), time=self._time)
        ts.data = self.data.copy()
        return ts


def delegate(cls, attr_name, method_name):
    def delegated(self, *vargs, **kwargs):
        a = getattr(self, attr_name)
        m = getattr(a, method_name)
        return m(*vargs, **kwargs)
    setattr(cls, method_name, delegated)

dunder_funcs = ['truediv', 'add', 'sub', 'mul', 'floordiv', 'mod', 'pow', 'lshift', 'rshift', 'and', 'xor', 'or', 'iadd', 'isub', 'imul', 'idiv', 'ifloordiv', 'imod', 'ipow', 'ilshift', 'irshift', 'iand', 'ixor', 'radd', 'rsub', 'rmul', 'rdiv', 'rfloordiv', 'rmod', 'rpow', 'rlshift', 'rand', 'rxor', 'ror', 'lt', 'le', 'ne', 'eq', 'gt', 'ge','len']

# proxy the magic methods to the numpy array
for name in dunder_funcs:
    name = f"__{name}__"
    delegate(TimeSeries, 'data', name)


from dataclasses import dataclass
from typing import Callable, Any, Dict
from .timeseries import TimeSeries


def generate_title(ts: TimeSeries) -> str:
    """Generate a title based on common tags"""
    # get common labels
    labels = ts._get_unique_keys()
    if len(labels) > 0:
        return ".".join([str(v) for k, v in labels.items()])
    return "TimeSeries"


def generate_xlabel(ts: TimeSeries) -> str:
    return "Value"


def generate_ylabel(ts: TimeSeries) -> str:
    return "Value"


def generate_legend_labels(ts: TimeSeries) -> Dict[str, str]:
    common_labels = ts._get_unique_keys().keys()
    labels = []
    for l in ts.labels:
        labels.append(".".join([str(v) for k, v in l.items() if k not in common_labels]))
    return labels


def generate_chart_options(ts: TimeSeries) -> Dict[str, str]:
    return {}


@dataclass
class Plot:

    title_func: Callable[[Any], str] = generate_title
    xlabel_func: Callable[[Any], str] = generate_xlabel
    ylabel_func: Callable[[Any], str] = generate_ylabel
    legend_labels_func: Callable[[Any], str] = generate_legend_labels
    chart_options_func: Callable[[Any], str] = generate_chart_options

    def line_plot(self, ts: TimeSeries, **kwargs):
        """ Plot all TimeSeries columns as a line """
        return NotImplementedError

    def dist_plot(self, ts: TimeSeries, **kwargs):
        """ Plot all TimeSeries columns as a distribution """
        return NotImplementedError

    def stacked_plot(self, ts: TimeSeries, **kwargs):
        """ Plot all TimeSeries columns as a stacked chart """
        return NotImplementedError

    def timebox_plot(self, ts: TimeSeries, **kwargs):
        """ Plot all TimeSeries columns as a time box chart """
        return NotImplementedError

    def correlogram_plot(self, ts: TimeSeries, **kwargs):
        """ Plot all auto-correlation chart """
        return NotImplementedError

    def text_plot(self, ts: TimeSeries, **kwargs):
        """ Plot a single value """
        return NotImplementedError

    def heatmap_plot(self, ts: TimeSeries, **kwargs):
        """ Plot all TimeSeries columns as a heatmap """
        raise NotImplementedError

    def pacf(self, ts: TimeSeries, **kwargs):
        """ Plot a pacf chart """
        raise NotImplementedError

    def acf(self, ts: TimeSeries, **kwargs):
        """ Plot a acf chart """
        raise NotImplementedError

    def qq_plot(self, ts: TimeSeries, **kwargs):
        """ Plot a qq plot """
        raise NotImplementedError

    def probability_plot(self, ts: TimeSeries, **kwargs):
        """ Plot a probablility plot """
        raise NotImplementedError

    def polar_plot(self, ts: TimeSeries, **kwargs):
        """ Plot a polar plot """
        raise NotImplementedError



import os
import time
import numbers
from .timeseries import TimeSeries
from .time import convert_string_to_seconds
from .ast import AST


class Query:
    """
    A query class that generates an execution plan. This class can be 
    easily extended by overriding the `execute_plan` plan method.
    """

    DEFAULT_START_OFFSET = int(os.environ.get("DEFAULT_START_OFFSET", "3600"))
    DEFAULT_RESOLUTION = int(os.environ.get("DEFAULT_RESOLUTION", "60"))

    def __init__(self, *args):
        """Override __init__"""
        self.groupings = None
        self.period = None
        self.filters = []
        self.data = None
        self.generators = [a for a in args]

    def __iter__(self):
        """Make Query object an iterable"""
        return self

    def __next__(self):
        """Nothing to iterate"""
        raise StopIteration()

    def __getitem__(self, items):
        """Have the ability to fetch data via the getitem method

        Example
        -------

        t1 = Query(x.mean for x in "test")[x:y:z]

        params
        ------

        items: slice
            start = start_time_offset in seconds (now - start_time_offset = start_time)
            stop = end_time_offset in seconds (now - end_time_offset = end_time)
            step = resolution in seconds
        """
        self.period = items
        return self.execute_plan()

    def execute_plan(self):
        """Execute the query against the backend

        In this case there is a simple return of an empty
        TimeSeries with the shape of the calculated time period.
        This should be overridden by a backend to process the query.

        """
        period = self._process_period()
        # should return time series
        # return empty timeseries based on the period
        timeindex = [
            x for x in range(period["start_time"], period["end_time"], period["resolution"])
        ]
        t1 = TimeSeries(shape=(len(timeindex), 1), time=timeindex)
        # t1[:] = [[i,np.nan] for i in timeindex]
        return t1

    def _process_period(self):
        """ This processes the period from the __getitem__() function

        Slice Examples:
        --------------

        [3600] start_time = now - 3600 seconds, end_time=now
        [3600:1800] start_time = now - 3600 seconds, end_time = now - 1800 seconds
        [3600:1800:300] start_time = now - 3600 seconds, end_time = now - 1800 seconds, resolution = 5 minutes

        String Examples
        ---------------
        ['15m'] start_time = now - 900 seconds

        date string options are:

        s - seconds
        m - minutes
        h - hours
        d - days
        w - weeks
        y - years

        Returns
        -------
        a period dictionary
        """
        start_offset = self.DEFAULT_START_OFFSET
        end_offset = 0
        resolution = self.DEFAULT_RESOLUTION
        now = int(time.time())
        if isinstance(self.period, int):
            start_offset = self.period
        elif isinstance(self.period, slice):
            start_offset = (
                self.period.start
                if self.period and self.period.start
                else self.DEFAULT_START_OFFSET
            )
            end_offset = self.period.stop if self.period and self.period.stop else 0
            resolution = (
                self.period.step if self.period and self.period.step else self.DEFAULT_RESOLUTION
            )
        elif isinstance(self.period, str):
            try:
                start_offset = convert_string_to_seconds(self.period)
            except:
                self._raise_period_index_error()
        else:
            self._raise_period_index_error()

        period = {
            "start_time": now - start_offset,
            "end_time": now - end_offset,
            "resolution": resolution,
        }
        return period

    def _raise_period_index_error(self):
        raise IndexError(
            f"Only slices, integers, and date format strings are allowed not {type(self.period)}"
        )

    def _generate_plan(self):
        """This function generates a plan for the query engine to execute

        Return Example
        --------------

        {
                "metrics": [
                    {
                        "eval": None,
                        "metrics": ["metric1"],
                        "filters": [],
                        "group": [['tag1'], 'mean'],
                        "vars": [{'name': 'x', 'labels': ['max']}]
                    },{
                        "eval": None,
                        "metrics": ["metric2"],
                        "filters": [],
                        "group": [['tag1'], 'mean'],
                        "vars": [{'name': 'x', 'labels': ['min']}]
                    }
                ],
                "eval": [['x', 'y', 'BINARY_SUBTRACT']],
                "filters": [],
                "group": None,
                "vars": [{'name': 'x', 'labels': []}, {'name': 'y', 'labels': []}]
        }
        """
        return AST(self.generators, self.groupings).decompile()

    def by(self, labels, func="mean"):
        """Adds a group by step to the Query plan
        
        Params:
        -------
        Labels: list of strings
            The labels that are to be grouped on
        func:  a ufunc
            The function to merge the different stream with

        Returns
        -------
        Self
        """
        if not isinstance(labels, list):
            labels = [labels]
        if not self.groupings:
            self.groupings = []
        self.groupings = [labels, func]
        return self

    def fetch(self):
        """Fetch all items.  This can be overridden with sane defaults"""
        return self[:]

    def range(self, start, end=None, resolution=None):
        """ 
        Fetch points by a specified time range 
        
        Params:
        -------
        start: int
            A unixtimestamp for the start
        end: int
            A unixtimestamp for the end
        resolution: int
            A int for the resolution

        Returns
        -------
        TimeSeries
        """
        now = int(time.time())
        end_offset = None if end == None else end - start
        res = self.DEFAULT_RESOLUTION if resolution is None else resolution

        s = slice(now - start, end_offset, res)
        return self[s]



import dis
import collections
import types


class Node:
    """
    A generic class for a tree Node.  

    params
    ------

    left: any
        the left node which will be evaluated before the right node
    right: any
        the right node evaluated after the left

    """

    left = None
    right = None

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def pprint(self, level):
        """ Pretty print for debugging purposes """
        nl = "\n"
        tabs = "\t" * level
        return f"{tabs}{self.__class__.__name__}:{nl}{self.left.pprint(level+1)},{nl}{self.right.pprint(level+1)}"

    def __eq__(self, v):
        """ Override the eq so tests can compare all children nodes as well """
        if isinstance(self, type(v)):
            return self.left == v.left and self.right == v.right
        return False

    def __str__(self):
        """ Pretty print instead of using string """
        return self.pprint(0)


class Filter(Node):
    """ Used when a filter is used in the generator """

    pass


class Group(Node):
    """ Used when a filter is used in the generator """

    pass


class CompareEqual(Node):
    """ Used when a filter is used in the generator """

    pass


class CompareNotEqual(Node):
    """ Used when a filter is used in the generator """

    pass


class CompareIn(Node):
    """ Used when a filter is used in the generator """

    pass


class CompareNotIn(Node):
    """ Used when a filter is used in the generator """

    pass


class LoadAttr(Node):
    """ Used when a filter is used in the generator """

    pass


class Func(Node):
    """ Used when a filter is used in the generator """

    pass


class FuncCall(Node):
    """ Used when a filter is used in the generator """

    pass


class FuncArgs(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryModulo(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryPower(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryFloorDivide(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryTrueDivide(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryMatrixMultiply(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryMultiply(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryAdd(Node):
    """ Used when a filter is used in the generator """

    pass


class BinarySubtract(Node):
    """ Used when a filter is used in the generator """

    pass


class BinaryXOR(Node):
    """ Used when a filter is used in the generator """

    pass


class Value:
    """ 
    A generic value node.  Using this node indicated the bottom of tree

    value: any
        The value to be stored. 
    """

    value = None

    def __init__(self, value):
        self.value = value

    def pprint(self, level):
        """ pretty print for debugging """
        tabs = "\t" * level
        val = self.value
        if isinstance(self.value, list):
            val = "[\n"
            for v in self.value:
                try:
                    val += getattr(v, "pprint")(level + 1)
                except:
                    val += tabs + "\t" + str(v)
                val += ",\n"
            val += f"{tabs}]"
        return f"{tabs}{self.__class__.__name__}: {val}"

    def __eq__(self, v):
        """ Used to compare Value objects """
        if type(v) == type(self):
            return self.value == v.value
        return False

    def __str__(self):
        return self.pprint(0)

    def __repr__(self):
        return f"Value<{self.value}>"


class Metric(Value):
    """ A class that represents the metric with the metric name as the value """

    pass


class AST:
    """
    A class to represent a generic abstract syntax to be consumed by a backend

    gen: generator(s)
        The generator(s) to be processed
    group_by: list
        A list of labels to group by
    working_var: str
        A temporary value to track which variable is currently "loaded"
    variables: map of nodes
        A map of nodes that represent the variables
    current_gen: generator
        The generator being processed
    stack: list of nodes/values
        The processing stack for the class
    """

    def __init__(self, gen, group_by=None):
        self.gen = gen
        self.group_by = group_by
        self.working_var = None
        self.variables = {}
        self.current_gen = None
        self.stack = []

    def decompile(self):
        """ The decompile function """
        if not isinstance(self.gen, list):
            self.gen = [self.gen]
        for gen in self.gen:
            self.current_gen = gen
            bytecode = dis.Bytecode(gen)
            for instr in bytecode:
                try:
                    op = self.__getattribute__("_" + instr.opname.lower())
                    op(instr)
                except:
                    raise NotImplementedError(f"opname of {instr.opname} is not supported")
            rtn = self.stack.pop()
            if self.group_by:
                rtn = Group(rtn, Value(self.group_by))
            self.stack.append(rtn)
        return rtn

    # op handlers

    def _binary_add(self, instr):
        self._binary_op(BinaryAdd)

    def _binary_floor_divide(self, instr):
        self._binary_op(BinaryFloorDivide)

    def _binary_matrix_multiply(self, instr):
        self._binary_op(BinaryMatrixMultiply)

    def _binary_multiply(self, instr):
        self._binary_op(BinaryMultiply)

    def _binary_modulo(self, instr):
        self._binary_op(BinaryModulo)

    def _binary_power(self, instr):
        self._binary_op(BinaryPower)

    def _binary_op(self, cl):
        right = self.stack.pop()
        left = self.stack.pop()
        self.stack.append(cl(left, right))

    def _binary_subscr(self, instr):
        index = self.stack.pop()
        target = self.stack.pop()

        self.stack.append(Value(target.value[index.value]))

    def _binary_subtract(self, instr):
        self._binary_op(BinarySubtract)

    def _binary_true_divide(self, instr):
        self._binary_op(BinaryTrueDivide)

    def _binary_xor(self, instr):
        self._binary_op(BinaryXOR)

    def _call_function(self, instr):
        kwargs = {}
        args = []
        arg_count = instr.arg
        if instr.opname == "CALL_FUNCTION_KW":
            kwarg_names = self.stack.pop()
            for n in reversed(kwarg_names.value):
                kwargs[n] = self.stack.pop().value
                arg_count -= 1
        for _ in range(arg_count):
            args.insert(0, self.stack.pop())
        name = self.stack.pop()
        func = FuncCall(name, FuncArgs(Value(args), Value(kwargs)))
        self.stack.append(func)

    def _call_function_kw(self, instr):
        self._call_function(instr)

    def _compare_op(self, instr):
        ops = {"in": CompareIn, "==": CompareEqual, "!=": CompareNotEqual, "not in": CompareNotIn}
        right = self.stack.pop()
        left = self.stack.pop()
        cl = ops[instr.argval](left.right, right)
        f = Filter(left.left, cl)
        self.stack.append(f)

    def _for_iter(self, instr):
        from .query import Query

        stack_val = self.stack.pop()
        if type(stack_val.value) == type(iter("")):
            self.stack.append(Metric(["".join(stack_val.value)][0]))
        else:
            try:
                for q in stack_val.value.generators:
                    a = AST(q).decompile()
                    self.stack.append(a)
                if stack_val.value.groupings:
                    self.group_by = stack_val.value.groupings
            except:  # not a query object
                for q in stack_val.value:
                    self.stack.append(Metric(q))


    def _jump_absolute(self, instr):
        pass

    def _load_attr(self, instr):
        self.stack.append(LoadAttr(self.stack.pop(), Value(instr.argval)))

    def _load_const(self, instr):
        if not instr.is_jump_target:
            self.stack.append(Value(instr.argval))

    def _load_deref(self, instr):
        self.stack.append(Value(self.current_gen.gi_frame.f_locals.get(instr.argval, "unknown")))

    def _load_fast(self, instr):
        if instr.argval.startswith("."):
            val = Value(self.current_gen.gi_frame.f_locals[instr.argval])
            self.stack.append(val)
            self.variables[instr.argval] = val
        else:
            self.working_var = instr.argval
            self.stack.append(self.variables[instr.argval])

    def _load_global(self, instr):
        self.stack.append(Value(self.current_gen.gi_frame.f_globals.get(instr.argval, instr.argval)))

    def _pop_jump_if_false(self, instr):
        self.variables[self.working_var] = self.stack.pop()
        self.working_var = None

    def _pop_top(self, instr):
        pass

    def _return_value(self, instr):
        pass

    def _store_fast(self, instr):
        self.variables[instr.argval] = self.stack.pop(0)

    def _unpack_sequence(self, instr):
        pass

    def _yield_value(self, instr):
        pass




#!/usr/bin/env python

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="timeseriesql",
    version="0.1.5",
    description="A Pythonic query language for time series data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/mbeale/timeseriesql",
    author="Michael Beale",
    author_email="michael.beale@gmail.com",
    license="MIT",
    packages=["timeseriesql", "timeseriesql.backends"],
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    install_requires=["numpy"],
    python_requires=">=3.6",
)




class TimeChunk:
    """ The TimeChuck class represents a mask over a NumpyArray """

    row_mask = None
    col_mask = None

    def __init__(self, row_mask, col_mask):
        """
        Set the intial row and column mask

        Parameters
        ----------

        row_mask : slice
            designates which rows to select
        col_mask : slice
            designates which columns to select
        """
        self.row_mask = row_mask
        self.col_mask = col_mask
